"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.claimCheck = exports.claimEquals = exports.claimIncludes = exports.requiredScopes = void 0;
const oauth2_bearer_1 = require("./oauth2-bearer");
const checkJSONPrimitive = (value) => {
    if (typeof value !== 'string' &&
        typeof value !== 'number' &&
        typeof value !== 'boolean' &&
        value !== null) {
        throw new TypeError("'expected' must be a string, number, boolean or null");
    }
};
const isClaimIncluded = (claim, expected) => (payload) => {
    if (!(claim in payload)) {
        throw new oauth2_bearer_1.InvalidTokenError(`Missing '${claim}' claim`);
    }
    let actual = payload[claim];
    if (typeof actual === 'string') {
        actual = actual.split(' ');
    }
    else if (!Array.isArray(actual)) {
        return false;
    }
    actual = new Set(actual);
    return expected.every(Set.prototype.has.bind(actual));
};
const requiredScopes = (scopes) => {
    if (typeof scopes === 'string') {
        scopes = scopes.split(' ');
    }
    else if (!Array.isArray(scopes)) {
        throw new TypeError("'scopes' must be a string or array of strings");
    }
    const fn = isClaimIncluded('scope', scopes);
    return (0, exports.claimCheck)((payload) => {
        if (!('scope' in payload)) {
            throw new oauth2_bearer_1.InsufficientScopeError(scopes, "Missing 'scope' claim");
        }
        if (!fn(payload)) {
            throw new oauth2_bearer_1.InsufficientScopeError(scopes);
        }
        return true;
    });
};
exports.requiredScopes = requiredScopes;
const claimIncludes = (claim, ...expected) => {
    if (typeof claim !== 'string') {
        throw new TypeError("'claim' must be a string");
    }
    expected.forEach(checkJSONPrimitive);
    return (0, exports.claimCheck)(isClaimIncluded(claim, expected), `Unexpected '${claim}' value`);
};
exports.claimIncludes = claimIncludes;
const claimEquals = (claim, expected) => {
    if (typeof claim !== 'string') {
        throw new TypeError("'claim' must be a string");
    }
    checkJSONPrimitive(expected);
    return (0, exports.claimCheck)((payload) => {
        if (!(claim in payload)) {
            throw new oauth2_bearer_1.InvalidTokenError(`Missing '${claim}' claim`);
        }
        return payload[claim] === expected;
    }, `Unexpected '${claim}' value`);
};
exports.claimEquals = claimEquals;
const claimCheck = (fn, errMsg) => {
    if (typeof fn !== 'function') {
        throw new TypeError("'claimCheck' expects a function");
    }
    return (payload) => {
        if (!payload) {
            throw new oauth2_bearer_1.UnauthorizedError();
        }
        if (!fn(payload)) {
            throw new oauth2_bearer_1.InvalidTokenError(errMsg);
        }
    };
};
exports.claimCheck = claimCheck;
