"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = require("http");
const url_1 = require("url");
const anyBody = require("body/any");
const got_1 = __importDefault(require("got"));
const typeis = require("type-is");
const get_token_1 = __importDefault(require("./get-token"));
const start = (server) => new Promise((resolve) => server.listen(0, () => resolve(`http://localhost:${server.address().port}`)));
const handler = (req, res) => {
    let query;
    new url_1.URL(req.url, 'http://localhost').searchParams.forEach((val, key) => {
        query = query || {};
        query[key] = val;
    });
    anyBody(req, res, (err, body) => {
        try {
            res.end((0, get_token_1.default)(req.headers, query, body, !!typeis.is(req.headers['content-type'], ['urlencoded'])));
        }
        catch (e) {
            res.statusCode = e.statusCode;
            res.statusMessage = e.message;
            res.end();
        }
    });
};
describe('get-token', () => {
    let server;
    let url;
    beforeEach(async () => {
        server = (0, http_1.createServer)(handler);
        url = await start(server);
    });
    afterEach((done) => {
        server.close(done);
    });
    it('should fail when there are no tokens', async () => {
        await expect((0, got_1.default)(url)).rejects.toThrowError('Response code 401 (Unauthorized)');
    });
    it('should get the token from the header', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            headers: {
                authorization: 'Bearer token',
            },
        })).resolves.toEqual('token');
    });
    it('should do case insensitive check for header', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            headers: {
                authorization: 'bearer token',
            },
        })).resolves.toEqual('token');
    });
    it('should fail for malformed header', async () => {
        await expect((0, got_1.default)(url, {
            headers: {
                authorization: 'foo token',
            },
        })).rejects.toThrowError('Response code 401 (Unauthorized)');
    });
    it('should fail for empty header', async () => {
        await expect((0, got_1.default)(url, {
            headers: {
                authorization: 'Bearer ',
            },
        })).rejects.toThrowError('Response code 401 (Unauthorized)');
    });
    it('should get the token from the query string', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            searchParams: { access_token: 'token' },
        })).resolves.toEqual('token');
    });
    it('should succeed to get the token from the query string for POST requests', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            method: 'POST',
            searchParams: { access_token: 'token' },
        })).resolves.toEqual('token');
    });
    it('should get the token from the request payload', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            method: 'POST',
            form: { access_token: 'token' },
        })).resolves.toEqual('token');
    });
    it('should fail to get the token from JSON request payload', async () => {
        await expect((0, got_1.default)(url, {
            method: 'POST',
            json: { access_token: 'token' },
        })).rejects.toThrowError('Response code 401 (Unauthorized)');
    });
    it('should succeed to get the token from request payload for GETs', async () => {
        await expect((0, got_1.default)(url, {
            resolveBodyOnly: true,
            allowGetBody: true,
            method: 'GET',
            form: { access_token: 'token' },
        })).resolves.toEqual('token');
    });
    it('should fail if more than one method is used', async () => {
        await expect((0, got_1.default)(url, {
            searchParams: { access_token: 'token' },
            headers: {
                authorization: 'Bearer token',
            },
        })).rejects.toThrowError('Response code 400 (More than one method used for authentication)');
    });
});
